@@ -97,6 +97,7 @@ typedef struct {
 } DiscoveryResponse;
 
 #define DEFAULT_MESSAGE_DELAY 20 
+#define VERSION_PATTERN "[0-9]+$"
 
 /* Function prototypes */
 static void
@@ -123,6 +124,8 @@ static void
 discovery_response_free         (DiscoveryResponse  *response);
 static gboolean
 process_queue                   (gpointer            data);
+static char *
+get_version_for_target          (char *target);
 static GRegex *
 create_target_regex             (const char         *target,
                                  GError            **error);
@@ -503,10 +506,15 @@ gssdp_resource_group_set_available (GSSDPResourceGroup *resource_group,
                 GMainContext *context;
                 int timeout;
 
-                /* We want to re-announce before actually timing out */
+                /* We want to re-announce at least 3 times before the resource
+                 * group expires to cope with the unrelialble nature of UDP.
+                 *
+                 * Read the paragraphs about 'CACHE-CONTROL' on pages 21-22 of
+                 * UPnP Device Architecture Document v1.1 for further details.
+                 * */
                 timeout = resource_group->priv->max_age;
-                if (timeout > 2)
-                        timeout = timeout / 2 - 1;
+                if (G_LIKELY (timeout > 6))
+                        timeout = (timeout / 3) - 1;
 
                 /* Add re-announcement timer */
                 resource_group->priv->timeout_src =
@@ -974,14 +982,67 @@ queue_message (GSSDPResourceGroup *resource_group,
 }
 
 /**
- * Send ssdp:alive message for @resource
+ * Add ssdp:alive messages for the given resource to sending queue
+ **/
+static void
+queue_alive_messages (GSSDPResourceGroup *resource_group,
+                      guint               max_age,
+                      const char         *location,
+                      const char         *al,
+                      const char         *server_id,
+                      const char         *target,
+                      const char         *usn)
+{
+        char *version_str;
+        char *message;
+
+        version_str = get_version_for_target ((char *) target);
+        if (version_str != NULL) {
+                char *_target;
+                int version;
+
+                version = atoi (version_str);
+                _target = g_strndup (target, version_str - target);
+
+                while (version > 0) {
+                        message = g_strdup_printf
+                                              (SSDP_ALIVE_MESSAGE_WITH_VERSION,
+                                               max_age,
+                                               location,
+                                               al,
+                                               server_id,
+                                               _target,
+                                               version,
+                                               usn);
+
+                        queue_message (resource_group, message);
+
+                        version--;
+                }
+
+                g_free (_target);
+        } else {
+                message = g_strdup_printf (SSDP_ALIVE_MESSAGE,
+                                           max_age,
+                                           location,
+                                           al,
+                                           server_id,
+                                           target,
+                                           usn);
+
+                queue_message (resource_group, message);
+        }
+}
+
+/**
+ * Send ssdp:alive messages for @resource
  **/
 static void
 resource_alive (Resource *resource)
 {
         GSSDPClient *client;
         guint max_age;
-        char *al, *message;
+        char *al;
 
         if (!resource->initial_alive_sent) {
                 /* Unannounce before first announce. This is done to
@@ -999,15 +1060,13 @@ resource_alive (Resource *resource)
 
         al = construct_al (resource);
 
-        message = g_strdup_printf (SSDP_ALIVE_MESSAGE,
-                                   max_age,
-                                   (char *) resource->locations->data,
-                                   al ? al : "",
-                                   gssdp_client_get_server_id (client),
-                                   resource->target,
-                                   resource->usn);
-
-        queue_message (resource->resource_group, message);
+        queue_alive_messages (resource->resource_group,
+                              max_age,
+                              (const char *) resource->locations->data,
+                              al ? al : "",
+                              gssdp_client_get_server_id (client),
+                              resource->target,
+                              resource->usn);
 
         g_free (al);
 }
@@ -1018,14 +1077,38 @@ resource_alive (Resource *resource)
 static void
 resource_byebye (Resource *resource)
 {
+        char *version_str;
         char *message;
 
-        /* Queue message */
-        message = g_strdup_printf (SSDP_BYEBYE_MESSAGE,
-                                   resource->target,
-                                   resource->usn);
-        
-        queue_message (resource->resource_group, message);
+        version_str = get_version_for_target (resource->target);
+        if (version_str != NULL) {
+                char *_target;
+                int version;
+
+                version = atoi (version_str);
+                _target = g_strndup (resource->target,
+                                     version_str - resource->target);
+
+                while (version > 0) {
+                        message = g_strdup_printf
+                                              (SSDP_BYEBYE_MESSAGE_WITH_VERSION,
+                                               _target,
+                                               version,
+                                               resource->usn);
+
+                        queue_message (resource->resource_group, message);
+
+                        version--;
+                }
+
+                g_free (_target);
+        } else {
+                message = g_strdup_printf (SSDP_BYEBYE_MESSAGE,
+                                           resource->target,
+                                           resource->usn);
+
+                queue_message (resource->resource_group, message);
+        }
 }
 
 /**
@@ -1055,29 +1138,39 @@ resource_free (Resource *resource)
         g_slice_free (Resource, resource);
 }
 
+/* Gets you the pointer to the version part in the target string */
+static char *
+get_version_for_target (char *target)
+{
+        char *version;
+
+        if (strncmp (target, "urn:", 4) != 0) {
+                /* target is not a URN so no version. */
+                return NULL;
+        }
+
+        version = g_strrstr (target, ":") + 1;
+        if (version == NULL ||
+            !g_regex_match_simple (VERSION_PATTERN, version, 0, 0))
+                return NULL;
+
+        return version;
+}
+
 static GRegex *
 create_target_regex (const char *target, GError **error)
 {
         GRegex *regex;
         char *pattern;
         char *version;
-        char *version_pattern;
 
-        if (strncmp (target, "urn:", 4) != 0) {
-                /* target is not a URN, No need to deal with version. */
-                return g_regex_new (target, 0, 0, error);
-        }
-
-        version_pattern = "[0-9]+$";
         /* Make sure we have enough room for version pattern */
         pattern = g_strndup (target,
-                             strlen (target) + strlen (version_pattern));
+                             strlen (target) + strlen (VERSION_PATTERN));
 
-        version = g_strrstr (pattern, ":") + 1;
-        if (version != NULL &&
-            g_regex_match_simple (version_pattern, version, 0, 0)) {
-                strcpy (version, version_pattern);
-        }
+        version = get_version_for_target (pattern);
+        if (version != NULL)
+                strcpy (version, VERSION_PATTERN);
 
         regex = g_regex_new (pattern, 0, 0, error);
 
